'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var serializeAsCode = require('serialize-as-code');

function _extends() {
  _extends = Object.assign || function (target) {
    for (var i = 1; i < arguments.length; i++) {
      var source = arguments[i];

      for (var key in source) {
        if (Object.prototype.hasOwnProperty.call(source, key)) {
          target[key] = source[key];
        }
      }
    }

    return target;
  };

  return _extends.apply(this, arguments);
}

const IGNORE = Symbol.for('__Spy_IGNORE__');
const serialize = serializeAsCode.Serializer.create(o => o === IGNORE && '>IGNORED<' || undefined);

const SPY_COMPARE_FAILED = 'Spy.COMPARE failed';
const SPY_MAPPER_FAILED = 'Spy.MAPPER failed';

class SpyComparator {
  constructor(comparator) {
    this._func = void 0;
    this._func = comparator;
  }

  compare(arg) {
    const result = this._func(arg);

    if (typeof result === 'string') return [`${SPY_MAPPER_FAILED} [${result}]`];
    if (result === false) return [SPY_COMPARE_FAILED];
    return undefined;
  }

}

const COMPARE = comparator => new SpyComparator(comparator);

const MAPPER = (from, to) => new SpyComparator(mapper => {
  const result = mapper(...(Array.isArray(from) ? from : [from]));
  const diff = differenceOf(result, to);
  return diff ? `${serialize(result)} did not match ${serialize(to)}: ${diff}` : undefined;
});

const __different = type => ['different ' + type];

const __diff = (a, b, useOwnEquals, alreadyComparedArray = []) => {
  if (a === IGNORE || b === IGNORE) return undefined;
  if (a instanceof SpyComparator) return a.compare(b);
  if (b instanceof SpyComparator) return b.compare(a);
  if (a === b) return undefined;
  if (a === undefined || b === undefined) return ['one was undefined'];
  if (a === null || b === null) return ['one was null'];
  const aClass = Object.prototype.toString.call(a);
  const bClass = Object.prototype.toString.call(b);
  if (aClass !== bClass) return __different(`object types: ${aClass} <-> ${bClass}`);

  switch (aClass) {
    case '[object RegExp]':
      if (String(a) === String(b)) {
        return undefined;
      }

      return __different('regexp');

    case '[object String]':
      return __different('string');

    case '[object Function]':
      return __different('function');

    case '[object AsyncFunction]':
      return __different('async function');

    case '[object Number]':
      if (isNaN(a) && isNaN(b)) {
        return undefined;
      }

      return __different('number');

    case '[object BigInt]':
      return __different('BigInt');

    case '[object Date]':
      if (Number(a) === Number(b)) {
        return undefined;
      }

      return __different('date');

    case '[object Boolean]':
      return __different('bool');

    case '[object Symbol]':
      return __different('symbols');

    case '[object Error]':
      if (String(a) === String(b)) {
        return undefined;
      }

      return __different('error');

    default:
      if (a.constructor !== b.constructor) {
        return __different('constructor');
      }

  }

  if (useOwnEquals && a.equals instanceof Function) {
    if (a.equals(b)) {
      return undefined;
    }

    return ['own equals method failed <- ' + 'Maybe you want to disable the usage ' + 'of own equals implementation? ' + '[ Use: spy.configure({useOwnEquals: false}) ]'];
  }

  if (alreadyComparedArray.indexOf(a) !== -1) {
    return undefined;
  }

  const compared = [...alreadyComparedArray, a];
  const keys = Array.from(new Set(Object.keys(a).concat(Object.keys(b))));

  for (const key of keys) {
    const diff = __diff(a[key], b[key], useOwnEquals, compared);

    if (diff !== undefined) {
      return [key, ...diff];
    }
  }

  return undefined;
};

const __serializeDifferentProp = (obj, diff) => {
  let toSerialize = obj;
  diff.slice(0, -1).forEach(key => {
    toSerialize = toSerialize[key];
  });
  return serialize(toSerialize);
};

const __diffToStr = (diff, info = '') => {
  if (diff.length === 1) return diff[0];
  return `--> ${diff[0]} / ` + diff.slice(1).join(' / ') + info;
};

const differenceOf = (a, b, config = {
  useOwnEquals: true
}) => {
  const diff = __diff(a, b, config.useOwnEquals);

  if (!diff) return undefined;

  const diffStr = __diffToStr(diff);

  if (diff.length < 2) return diffStr;

  const diffProp1 = __serializeDifferentProp(a, diff);

  const lastPart = diff[diff.length - 1];
  if (lastPart.indexOf(SPY_MAPPER_FAILED) === 0) return diffStr;
  const info = lastPart === SPY_COMPARE_FAILED ? `called with: ${diffProp1}` : `${diffProp1} != ${__serializeDifferentProp(b, diff)}`;
  return `${diffStr} [${info}]`;
};

const toError = (msgOrError, spyName) => msgOrError instanceof Error ? msgOrError : new Error(msgOrError || `${spyName} was requested to throw`);

const restoreAttributeForEntry = value => {
  const {
    obj,
    method,
    methodName
  } = value;

  if (obj) {
    obj[methodName] = method;
  }
};

class SpyRegistry {
  constructor() {
    this.register = void 0;
    this.persReg = void 0;
    this.registerCount = void 0;
    this.register = {};
    this.persReg = {};
    this.registerCount = 0;
  }

  restoreAll() {
    Object.values(this.register).forEach(entry => {
      restoreAttributeForEntry(entry);
    });
    this.register = {};
  }

  restore(index) {
    const entry = this.register[index];

    if (entry) {
      restoreAttributeForEntry(entry);
      delete this.register[index];
    }
  }

  push(obj, methodName) {
    this.registerCount += 1;
    this.register[this.registerCount] = {
      obj,
      method: obj[methodName],
      methodName
    };
    return this.registerCount;
  }

  getOriginalMethod(index) {
    const entry = this.register[index];

    if (entry) {
      return entry.method;
    }
  }

  persist(index, intoPersReg) {
    const fromReg = intoPersReg ? this.register : this.persReg;
    const toReg = intoPersReg ? this.persReg : this.register;
    const entry = fromReg[index];

    if (entry) {
      toReg[index] = entry;
      delete fromReg[index];
    }
  }

}

const uninitialized = method => () => {
  throw new Error(`Method '${String(method)}' was not initialized on Mock.`);
};

const defaultScope = Symbol('__Spy_global__');
const _mocks = {
  [defaultScope]: []
};
let scope = defaultScope;
const setScope = scoping => {
  if (scoping) {
    _mocks[scoping] = [];
    scope = scoping;
  } else scope = defaultScope;
};

const registerMock = (mocked, returns) => {
  const mock = {};

  _mocks[scope].push({
    mocked,
    mock,
    scope,
    returns
  });

  return mock;
};

const createMock = (obj, methods, returns) => {
  const mock = registerMock(obj, returns);
  methods.forEach(method => {
    mock[method] = uninitialized(method);
  });
  return mock;
};

const couldNotInitError = (scope, additional) => new Error(`Could not initialize mock for ${scope === defaultScope ? 'global scope' : `scope "${scope}"`}, because:\n${additional}`);

const initMock = ({
  mocked,
  mock,
  scope,
  returns
}, spyOn) => {
  Object.keys(mock).forEach(method => {
    try {
      mock[method] = spyOn(mocked, method).returns(returns);
    } catch (e) {
      throw couldNotInitError(scope, e.message);
    }
  });
};

const initMockScope = (scoping, spyOn) => {
  Object.values(_mocks[scoping]).forEach(mock => initMock(mock, spyOn));
};

const initMocks = (spyOn, scoping) => {
  initMockScope(defaultScope, spyOn);
  scoping && initMockScope(scoping, spyOn);
};

const pathDirname = p => require('path').dirname(p);

const pathJoin = (...p) => require('path').join(...p);

const fsExistsSync = p => require('fs').existsSync(p);

const _testSuite = {
  isJest: !!jest,
  isCJS: !!require
};
const runner = {};
const oldDescribe = describe;

const newDescribe = (name, suite) => {
  oldDescribe(name, () => {
    const scoping = String(name);
    setScope(scoping);
    beforeEach(() => {
      runner.beforeEach && runner.beforeEach(scoping);
    });
    afterEach(() => {
      runner.afterEach && runner.afterEach(scoping);
    });
    const rv = suite();
    setScope(undefined);
    return rv;
  });
};

newDescribe.each = oldDescribe.each;
newDescribe.only = oldDescribe.only;
newDescribe.skip = oldDescribe.skip;
describe = newDescribe;

const configure = other => {
  if (other.afterEach) runner.afterEach = other.afterEach;
  if (other.beforeEach) runner.beforeEach = other.beforeEach;
};

const addSnapshotSerializer = serializer => {
  _testSuite.isJest && expect && expect.addSnapshotSerializer && expect.addSnapshotSerializer(serializer);
};

const STACK_NUM_CREATE_MOCK = 5;

const __caller = () => {
  const traceFn = Error.prepareStackTrace;

  Error.prepareStackTrace = (_error, stack) => stack;

  const stack = new Error().stack;
  Error.prepareStackTrace = traceFn;
  return stack[STACK_NUM_CREATE_MOCK].getFileName();
};

const __callerBasedir = () => pathDirname(__caller());

const __getAbsolutePath = moduleName => pathJoin(__callerBasedir(), moduleName);

const __getDepDir = (moduleName, file) => moduleName.includes('/') ? pathDirname(file) : file;

const __getNodeModulePath = moduleName => {
  let baseDir = __callerBasedir();

  const calcFile = () => pathJoin(baseDir, 'node_modules', moduleName);

  let file = calcFile();

  while (!fsExistsSync(__getDepDir(moduleName, file))) {
    if (baseDir === '/') {
      throw new Error(`spy4js: Could not find given module: "${moduleName}"`);
    }

    baseDir = pathJoin(baseDir, '..');
    file = calcFile();
  }

  return file;
};

const createModuleMock = (moduleName, names, returns) => {
  if (!_testSuite.isCJS) throw new Error('spy4js: Mocking a module only works if your test runner executes with CommonJS');
  const isNodeModule = moduleName.indexOf('.') !== 0 && moduleName.indexOf('/') !== 0;
  return createMock(require(isNodeModule ? __getNodeModulePath(moduleName) : __getAbsolutePath(moduleName)), names, returns);
};

const TestSuite = {
  addSnapshotSerializer,
  createModuleMock,
  configure
};

const CallOrder = {
  Idx: 0,
  lastCheck: 0
};
const registry = new SpyRegistry();
let __LOCK__ = true;
const symbolName = Symbol('__Spy_name__');
const symbolSnap = Symbol('__Spy_snap__');
const symbolIsSpy = Symbol('__Spy_isSpy__');
const symbolFunc = Symbol('__Spy_func__');
const symbolCalls = Symbol('__Spy_calls__');
const symbolConfig = Symbol('__Spy_config__');
const symbolIndex = Symbol('__Spy_index__');
const Symbols = {
  name: symbolName,
  snap: symbolSnap,
  isSpy: symbolIsSpy,
  func: symbolFunc,
  calls: symbolCalls,
  config: symbolConfig,
  index: symbolIndex
};
TestSuite.addSnapshotSerializer({
  test: v => v && v[Symbols.isSpy],
  print: spy => spy[Symbols.snap]
});
const DefaultSettings = {
  useOwnEquals: true,
  enforceOrder: false
};
const SpyHelperFunctions = {
  getCalls: spy => {
    const allCalls = spy[Symbols.calls];
    if (!DefaultSettings.enforceOrder) return allCalls;
    return allCalls.filter(({
      order
    }) => order > CallOrder.lastCheck);
  }
};
const SpyFunctions = {
  configure(config) {
    if (config.useOwnEquals !== undefined) {
      this[Symbols.config].useOwnEquals = config.useOwnEquals;
    }

    if (config.persistent !== undefined) {
      if (!this[Symbols.index]) {
        throw new Error(`\n\n${this[Symbols.name]} can not` + ' be configured to be persistent!' + ' It does not mock any object.');
      }

      this[Symbols.config].persistent = config.persistent;
      registry.persist(this[Symbols.index], this[Symbols.config].persistent);
    }

    return this;
  },

  calls(...funcs) {
    if (funcs.length === 0) {
      this[Symbols.func] = () => {};

      return this;
    }

    const max = funcs.length - 1;
    let counter = -1;

    this[Symbols.func] = (...args) => {
      counter++;
      return funcs[max < counter ? max : counter](...args);
    };

    return this;
  },

  returns(...args) {
    return this.calls(...args.map(arg => () => arg));
  },

  resolves(...args) {
    return this.returns(...(args.length ? args : [undefined]).map(arg => Promise.resolve(arg)));
  },

  rejects(...msgOrErrors) {
    return this.calls(...(msgOrErrors.length ? msgOrErrors : [undefined]).map(msgOrError => () => Promise.reject(toError(msgOrError, this[Symbols.name]))));
  },

  throws(msgOrError) {
    this[Symbols.func] = () => {
      throw toError(msgOrError, this[Symbols.name]);
    };

    return this;
  },

  reset() {
    this[Symbols.calls] = [];
    return this;
  },

  restore() {
    if (this[Symbols.config].persistent) {
      throw new Error(`\n\n${this[Symbols.name]} can not be restored!` + ' It was configured to be persistent.');
    }

    registry.restore(this[Symbols.index]);
    return this;
  },

  transparent() {
    return this.transparentAfter(0);
  },

  transparentAfter(callCount) {
    const oldFunc = this[Symbols.func];

    this[Symbols.func] = (...args) => {
      if (this[Symbols.calls].length > callCount) {
        const originalMethod = registry.getOriginalMethod(this[Symbols.index]);

        if (originalMethod) {
          return originalMethod(...args);
        }

        return;
      }

      return oldFunc(...args);
    };

    return this;
  },

  wasCalled(callCount) {
    const calls = SpyHelperFunctions.getCalls(this);
    const madeCalls = calls.length;

    if (callCount !== undefined) {
      if (madeCalls !== callCount) {
        throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls} times,` + ` but there were expected ${callCount} calls.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
      }
    } else if (madeCalls === 0) {
      if (this[Symbols.calls].length) {
        throw new Error(`\n\n${this[Symbols.name]} was not called,` + ` but there was expected at least one call.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
      } else {
        throw new Error(`\n\n${this[Symbols.name]} was never called!\n\n`);
      }
    }

    if (callCount !== 0) {
      const checked = calls[(callCount || 1) - 1];
      CallOrder.lastCheck = checked.order;
    }
  },

  hasCallHistory(...callHistory) {
    const calls = SpyHelperFunctions.getCalls(this);
    const madeCalls = calls.length;
    const callCount = callHistory.length;

    if (madeCalls !== callCount) {
      throw new Error(`\n\n${this[Symbols.name]} was called ${madeCalls} times,` + ` but the expected call history includes exactly ${callHistory.length} calls.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
    }

    if (callCount === 0) {
      throw new Error("\n\nPlease use 'wasNotCalled' instead of calling 'hasCallHistory' without any arguments");
    }

    const modifiedCallHistory = callHistory.map(arg => Array.isArray(arg) ? arg : [arg]);
    let hasErrors = false;
    const diffInfo = calls.map((call, index) => {
      const diff = differenceOf(call.args, modifiedCallHistory[index], this[Symbols.config]);
      if (diff) hasErrors = true;
      return diff;
    });
    if (hasErrors) throw new Error(`\n\n${this[Symbols.name]} was considered` + ' to be called with the following arguments in the given order:\n\n' + `${modifiedCallHistory.map((entry, index) => `call ${index}: ${serialize(entry)}`).join('\n')}\n\n` + 'Actually there were:\n\n' + this.showCallArguments(diffInfo));
    CallOrder.lastCheck = calls[callCount - 1].order;
  },

  wasNotCalled() {
    const calls = SpyHelperFunctions.getCalls(this);
    const madeCalls = calls.length;

    if (madeCalls !== 0) {
      throw new Error(`\n\n${this[Symbols.name]} was not considered to be called, but was called ${madeCalls} times.\n\n` + 'Actually there were:\n\n' + this.showCallArguments());
    }
  },

  wasCalledWith(...args) {
    const calls = SpyHelperFunctions.getCalls(this);
    const madeCalls = calls.length;
    const diffInfo = [];

    for (let i = 0; i < madeCalls; i++) {
      const checked = calls[i];
      const diff = differenceOf(checked.args, args, this[Symbols.config]);

      if (!diff) {
        CallOrder.lastCheck = checked.order;
        return;
      }

      diffInfo.push(diff);
    }

    throw new Error(`\n\n${this[Symbols.name]} was considered` + ' to be called with the following arguments:\n\n' + `    --> ${serialize(args)}\n\n` + 'Actually there were:\n\n' + this.showCallArguments(diffInfo));
  },

  wasNotCalledWith(...args) {
    let errorOccurred = false;

    try {
      this.wasCalledWith(...args);
    } catch (e) {
      errorOccurred = true;
    }

    if (!errorOccurred) {
      throw new Error(`\n\n${this[Symbols.name]} was called` + ' unexpectedly with the following arguments:\n\n' + `    --> ${serialize(args)}\n\n`);
    }
  },

  getAllCallArguments() {
    return this[Symbols.calls].map(({
      args
    }) => args);
  },

  getCallArguments(callNr = 0) {
    const madeCalls = this.getAllCallArguments();

    if (callNr % 1 !== 0 || callNr >= madeCalls.length) {
      throw new Error(`\n\nThe provided callNr "${callNr}" was not valid.\n\n` + `Made calls for ${this[Symbols.name]}:\n\n` + this.showCallArguments());
    }

    return madeCalls[callNr];
  },

  getCallArgument(callNr = 0, argNr = 0) {
    return this.getCallArguments(callNr)[argNr];
  },

  getCallCount() {
    return this.getAllCallArguments().length;
  },

  showCallArguments(additionalInformation = []) {
    const allMadeCalls = this[Symbols.calls];
    const madeCalls = SpyHelperFunctions.getCalls(this);
    const earlierCalls = allMadeCalls.length - madeCalls.length;
    const usedAdditionalInfo = new Array(earlierCalls).fill('!!! was called earlier (#enforceOrder)').concat(additionalInformation);

    if (allMadeCalls.length === 0) {
      return `${this[Symbols.name]} was never called!\n`;
    }

    let response = '';

    for (let i = 0; i < allMadeCalls.length; i++) {
      const args = serialize(allMadeCalls[i].args);
      response += `call ${i - earlierCalls}: ${args}\n`;

      if (usedAdditionalInfo[i]) {
        response += `        ${usedAdditionalInfo[i]}\n`;
      }
    }

    return response;
  }

};
const AllCreatedSpies = [];

const _createSpy = (name = '', __mock) => {
  const spy = function spy(...args) {
    spy[Symbols.calls].push({
      args,
      order: ++CallOrder.Idx
    });
    return spy[Symbols.func](...args);
  };

  if (__mock && !__LOCK__) {
    spy[Symbols.index] = registry.push(__mock.obj, __mock.methodName);
    spy[Symbols.name] = `the spy on '${name}'`;
    spy[Symbols.snap] = `Spy.on(${name})`;
  } else {
    spy[Symbols.index] = 0;
    spy[Symbols.name] = name || 'the spy';
    spy[Symbols.snap] = `Spy(${name})`;
  }

  spy[Symbols.isSpy] = true;

  spy[Symbols.func] = () => {};

  spy[Symbols.calls] = [];
  spy[Symbols.config] = _extends({}, DefaultSettings);
  Object.entries(SpyFunctions).forEach(([key, value]) => {
    spy[key] = value;
  });
  AllCreatedSpies.push(spy);
  return spy;
};

const Spy = name => _createSpy(name);

Spy.configure = config => {
  if (config.useOwnEquals !== undefined) {
    DefaultSettings.useOwnEquals = config.useOwnEquals;
  }

  if (config.enforceOrder !== undefined) {
    DefaultSettings.enforceOrder = config.enforceOrder;
  }

  TestSuite.configure({
    afterEach: config.afterEach,
    beforeEach: config.beforeEach
  });
};

Spy.IGNORE = IGNORE;
Spy.COMPARE = COMPARE;
Spy.MAPPER = MAPPER;

Spy.on = (obj, methodName) => {
  const method = obj[methodName];

  if (!(typeof method === 'function')) {
    throw new Error(`The object attribute '${methodName}' ` + `was: ${serialize(method)}\n\n` + 'You should only spy on functions!');
  }

  if (method[Symbols.isSpy]) {
    throw new Error(`The objects attribute '${methodName}'` + ' was already spied. Please make sure to spy' + ' only once at a time at any attribute.');
  }

  __LOCK__ = false;

  const spy = _createSpy(methodName, {
    obj,
    methodName
  });

  __LOCK__ = true;
  obj[methodName] = spy;
  return spy;
};

Spy.mock = (obj, ...methodNames) => createMock(obj, methodNames);

Spy.mockModule = (moduleName, ...methodNames) => TestSuite.createModuleMock(moduleName, methodNames);

Spy.mockReactComponents = (moduleName, ...methodNames) => TestSuite.createModuleMock(moduleName, methodNames, null);

Spy.initMocks = scope => {
  initMocks(Spy.on, scope);
};

Spy.restoreAll = () => {
  registry.restoreAll();
};

Spy.resetAll = () => {
  AllCreatedSpies.forEach(spy => spy.reset());
};

const defaultHooks = {
  beforeEach: Spy.initMocks,
  afterEach: () => {
    Spy.restoreAll();
    Spy.resetAll();
  }
};
TestSuite.configure(defaultHooks);

exports.Spy = Spy;
